PLUGIN_SOURCES = {
    'auth': 'spaceone.identity.plugin.auth.skeleton',
    'asset': 'spaceone.inventory.plugin.collector.skeleton',
    'metric': 'spaceone.monitoring.plugin.metric.skeleton',
    'log': 'spaceone.monitoring.plugin.log.skeleton',
    'webhook': 'spaceone.monitoring.plugin.webhook.skeleton',
    'cost': 'spaceone.cost_analysis.plugin.data_source.skeleton',
    'notification': 'spaceone.notification.plugin.protocol.skeleton',
}
